/*
 * Copyright (C) 2007 by Mark Pustjens <pustjens@dds.nl>
 *
 * This file is part of Binary Clock. 
 * Visit <http://unkie.org/wiki/misc/binary_clock>
 * for more information.
 *
 * Binary Clock Firmware is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Binary Clock Firmware is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * */

#include	<pic.h>
#include	<htc.h>

/*******************************
 *       global settings       *
 *******************************/

/* change the following to 1 if you want to drive the
 * basic version of the binary clock
 * */
#define BASIC 0
/* if you don't know what revision you are using, you should
 * set this to 1.
 * */
#define REVISION 0

/* the following defines the display blink speed
 * when changing the time
 */
#define BLINKS_PER_SEC	8	// blink 8 times per sec

/* It is recommended you use at least a 10Mhz crystal. With
 * a slower crystal dimming will not work as good.
 *
 * The correct value for the following setting can be calculated
 * by deviding the crystal frequency in hertz by 4.
 *
 * 1 cpu instruction cyle = 4 crystal oscillator periods.
 * with eg. a 4Mhz clock each second has 1000000 timer 0 ticks
 * per second.
 * */

#define T0_TICKS_PER_SECOND	3383375 // correct setting for 13.5335 Mhz crystal.

/*******************************
 *     end global settings     *
 *******************************/

#define TIME_TICKS	T0_TICKS_PER_SECOND
#define BLINK_TICKS	(T0_TICKS_PER_SECOND/BLINKS_PER_SEC)

/* timer settings */
near unsigned long int time_ticks = TIME_TICKS;
near unsigned long int blink_ticks = BLINK_TICKS;
#define OVERFLOW_TICKS 65536

/* we want to disable code protection,
 * disable the watchdog timer
 * and define we use a high speed crystal
 * oscillator 
 */
__CONFIG(UNPROTECT & WDTDIS & HS);

/* used for blinking the display */
bit display_on;

/* which ports are used for buttons to change settings */
#define BTNMODE	RB0	// change mode
#define BTNSET	RB7	// change current setting

/* contents of the matrix led display */
unsigned char matrix[6];
/* shorthand notations to ease typing */
#define s	matrix[5]
#define s10	matrix[4]
#define m	matrix[3]
#define m10	matrix[2]
#define h	matrix[1]
#define h10	matrix[0]

/* different modes of the clock nicely enumerated */
typedef enum {SETHR, SETMIN, SETLUM, DISPLAYOFF, TIME} modes;

/************************
 * interrupt routine to *
 * increment time       *
 ************************/

void interrupt timer0(void)
{
	unsigned char hour;
	/* clear timer interrupt, this is done here to 
	 * prevent drift which would occur if it was
	 * reset *after* running this function.
	 * note that this function must take less
	 * than T0_TICKS instructions.
	 *
	 * (normal instructions count as 1, jumps as 2
	 * see the chips manual for details)
	 */
	T0IF = 0; 

	/* time keeping */
	if (time_ticks < OVERFLOW_TICKS) {
		time_ticks += (TIME_TICKS - OVERFLOW_TICKS);
		/* one second passed, increment time */
		if (s < 9) {
			s++;
		} else {
			s = 0;
			if (s10 < 5) {
				s10++;
			} else {
				s10 = 0;
				if (m < 9) {
					m++;
				} else {
					m = 0;
					if (m10 < 5) {
						m10++;
					} else {
						m10 = 0;
						hour = (h10 * 4) + h;
						if (hour < 23) {
							hour++;
						} else {
							hour = 0;
						}
						h = hour % 10;
						h10 = hour / 10;
					}
				}
			}
		}
	} else {
		time_ticks -= OVERFLOW_TICKS;
	}

	/* display blinking */
	if (blink_ticks < OVERFLOW_TICKS) {
		blink_ticks += (BLINK_TICKS - OVERFLOW_TICKS);
		if (display_on == 1) display_on = 0;
		else display_on = 1;
	} else {
		blink_ticks -= OVERFLOW_TICKS;
	}
}

/* to detect button presses */
bit btnmode_detect = 0, btnset_detect = 0;

main(void)
{
	/* to generate the pwm to have the desired luminance */
	unsigned char lum_cnt = 0, lum = 0;
	/* to detect button presses */
	unsigned char btnmode_cnt = 0, btnset_cnt = 0;
	/* when turned on, blink to tell the user the time has to be set */
	modes mode = SETHR;
	/* for led matrix drawing */
	unsigned char x, y, xi, porta, i;

	/* setup timer0 */
	OPTION = 0xF7;	// select prescaler for timer0, set prescaler to 256
	T0CS = 0;	// select internal clock (instead of RA4)
	T0IE = 1;	// enable timer0 interrupt
	GIE = 1;	// enable interrupts

	/* initialize the matrix at 12:00:00 */
	matrix[0] = 0x01;
	matrix[1] = 0x02;
	matrix[2] = 0x00;
	matrix[3] = 0x00;
	matrix[4] = 0x00;
	matrix[5] = 0x00;

	/* initial pin vars */
	TRISB = 0x81;	/* pins 0 and 7 input, rest output */
	TRISA = 0x00;	/* output */
	/* set all output col pins to 0 (RB1..6 are col pins) */
	PORTB = 0x00;
	PORTA = 0x00;

	/* loop to infinity! hurray! */
	for(;;) {
		/* pwm for luminance */
		lum_cnt++;//	display_on = 1;
		/***************************
		 * detect button presses   *
		 * with debounce (bouncy!) *
		 ***************************/

		/* change mode button */
		if (BTNMODE == 0) {
			btnmode_cnt++;
			if (btnmode_cnt == 5) {
				if (!btnmode_detect) {
					/* mode button was pressed */
					btnmode_detect = 1;
					if (mode != TIME) {
						mode++;
					} else {
						mode = SETHR;
					}
				}
			}
		} else {
			btnmode_cnt = 0;
			btnmode_detect = 0;
		}

		/* change setting button */
		if (BTNSET == 0) {
			btnset_cnt++;
			if (btnset_cnt == 5) {
				if (!btnset_detect) {
					/* set button was pressed */
					btnset_detect = 1;
					switch (mode) {
					case SETMIN:
						/* change the minute of the day */
						if (m < 9) {
							m++;
						} else {
							m = 0;
							if (m10 < 5) {
								m10++;
							} else {
								m10 = 0;
							}
						}
						break;
					case SETHR:
						/* change the hour of the day */
						if ((h < 9 && h10 < 2) || (h < 3 && h10 == 2)) {
							h++;
						} else {
							h = 0;
							if (h10 < 1) {
								h10++;
							} else {
								h10 = 0;
							}	
						}
						break;
					case SETLUM:
						/* change the display luminance in 8 steps */
						if (lum != 0xFF) {
							lum = (lum << 1) | 1;
						} else {
							lum = 0;
						}
					case DISPLAYOFF:
					case TIME:
						/* do nothing */
						break;
					}
				}
			}
		} else {
			btnset_cnt = 0;
			btnset_detect = 0;
		}

		/*******************************
		 * draw the matrix led display *
		 *******************************/

		for (x = 0x02, xi = 0; x != 0x80; x <<= 1, xi++) { /* col pins 1..7 */
			for (y = 0x01; y != 0x10; y <<= 1) { /* row pins 0..3 */		
				if ((lum_cnt & lum) == lum) {
                                        switch (mode) {
                                        case SETHR:
                                                if ((x & 0x06) && display_on == 0) {
                                                        porta = 0x00;
                                                } else {
                                                        porta = y & matrix[xi];
                                                }
                                                break;
                                        case SETMIN:
                                                if ((x & 0x18) && display_on == 0) {
                                                        porta = 0x00;
                                                } else {
                                                        porta = y & matrix[xi];
                                                }
                                                break;
                                        case SETLUM:
                                                porta = y;
                                                break;
					case DISPLAYOFF:
						porta = 0x00;
						break;
                                        case TIME:
                                        default:
                                                porta = y & matrix[xi];
                                        }
                                } else {
                                        porta = 0x00;
                                }

				/* use the following for the basic version 
				 * where the leds are directly connected
				 * to the microcontroller pins.
				 */
				#if BASIC

				/* set row pin as output, rest as input */
				TRISA = ~y;
				/* set col pin and RB0,7 as input, rest as output 
				 * RB0 and RB7 are used as buttons, so must be input
				 */
				TRISB = ~x | 0x81;
				/* set the new value */
				PORTA = porta;

				/* use the following for a matrix using some transistors */
				#else

				#if REVISION 
				PORTA = 0x00;
				PORTB = 0x00;

				/* Give the transistors time to block completely 
				 * this fixed a bug relating to that for me.
				 * Unfortunately, this is compiler dependant.
				 */
				for (i = 0; i < 2; i++);

				TRISA = ~y;
				TRISB = ~x | 0x81;

				PORTA = porta;
				PORTB = x;

				#else
				/* In revision 1.0 the row pins must be
				 * driven inverted
				 */

				PORTA = 0xFF;
				PORTB = 0x00;

				for (i = 0; i < 2; i++);

				TRISA =0x00;// y;
				TRISB = ~x | 0x81;

				PORTA = ~porta;
				PORTB = x;

				#endif /* REVISION */
				#endif /* BASIC*/
			}
		}
	}
}
